from datetime import datetime
from time import sleep
from zoneinfo import ZoneInfo

from ..constants import POLLING_INTERVAL, RUN_WITH_LOGS
from ..helpers.web import (
    create_browser_session,
    get_reservations_html,
    InvalidCredentialsError,
)
from ..helpers.calendar_entries import (
    extract_calendar_entries,
    find_available_slots,
    write_calendar_entries_to_db,
)
from ..helpers.telegram_bot import notify_to_telegram


def run_once(email, password, token, chat_id, session=None, retries=0):
    if retries > 0:
        if RUN_WITH_LOGS:
            print(f"Sleep {2**retries} seconds before retrying")
        sleep(2**retries)
    browser_session = create_browser_session() if session is None else session
    try:
        html = get_reservations_html(
            session=browser_session, email=email, password=password
        )
    # TODO: Narrow down error class
    except Exception as e:
        if isinstance(e, InvalidCredentialsError):
            raise e
        # Crete new browser session when retrying to avoid any other type of error (like RemoteDisconnected errors)
        return run_once(
            email=email,
            password=password,
            token=token,
            chat_id=chat_id,
            session=None,
            retries=retries + 1,
        )

    calendar_entries = extract_calendar_entries(html)
    slots = find_available_slots(calendar_entries)
    if slots is not None:
        notify_to_telegram(available_slots=slots, token=token, chat_id=chat_id)
    write_calendar_entries_to_db(calendar_entries)

    return browser_session


# Polling should be run only between 7am and midnight
def run_once_conditionally(email, password, token, chat_id, session=None):
    now = datetime.now(ZoneInfo("Europe/Zurich"))
    if now.hour >= 7:
        return run_once(
            email=email,
            password=password,
            token=token,
            chat_id=chat_id,
            session=session,
        )
    else:
        return session


def run_in_loop(email, password, token, chat_id):
    browser_session = run_once_conditionally(
        email=email, password=password, token=token, chat_id=chat_id
    )
    sleep(POLLING_INTERVAL)
    while True:
        run_once_conditionally(
            session=browser_session,
            email=email,
            password=password,
            token=token,
            chat_id=chat_id,
        )
        sleep(POLLING_INTERVAL)



from datetime import datetime

from requests import Session

from ..constants import RUN_WITH_LOGS, TIMEOUT_GET_REQUEST

HOME_PAGE = "https://lakeridersclub.ch/index.php"
AUTHENTICATION_PAGE = "https://lakeridersclub.ch/membres/connexion.php"
CALENDAR_PAGE = "https://lakeridersclub.ch/membres/reservations.php"


class InvalidCredentialsError(Exception):
    pass


def create_browser_session():
    if RUN_WITH_LOGS:
        print("Creating browser session")
    session = Session()
    session.get(HOME_PAGE)
    return session


def get_session_authorized(session, email, password):
    if RUN_WITH_LOGS:
        print("Authenticating")

    session.post(
        url=AUTHENTICATION_PAGE,
        data={
            "adresse_electronique": email,
            "mot_de_passe": password,
            "rester_connecte": 1,
            "action": "se_connecter",
        },
    )


def _print_crawling_log():
    if RUN_WITH_LOGS:
        message = (
            f"Crawling lakeriders calendar on {datetime.now().strftime('%d %b, %H:%M')}"
        )
        print(f"{'-'*len(message)}\n{message}\n...\n..\n.")


def get_reservations_html(session, email, password):
    response = session.get(CALENDAR_PAGE, timeout=TIMEOUT_GET_REQUEST)
    if response.url == HOME_PAGE:
        # The session is not authenticated. Re-authenticate
        get_session_authorized(session=session, email=email, password=password)
        authenticated_response = session.get(CALENDAR_PAGE, timeout=TIMEOUT_GET_REQUEST)
        if authenticated_response.url == HOME_PAGE:
            raise InvalidCredentialsError("Invalid credentials")
        _print_crawling_log()
        return authenticated_response.text
    else:
        _print_crawling_log()
        return response.text



html = """
<!DOCTYPE html>
<html>
	<head>
	
		<meta charset="UTF-8">

<title>Lake Riders Club</title>

<meta name="description" xml:lang="fr" content="Lake Riders Club, club dédié au sport nautique sur le lac Léman" lang="fr" />
<meta name="keywords" xml:lang="fr" content="wakeboards, wakesurf, wakeskate, Sport nautique, Lac Léman, Genève, Suisse, Lake Riders Club" lang="fr" />

<meta name="viewport" content="initial-scale=1.0" />

<link rel="shortcut icon" href="/favicon.ico" />	
<link rel="stylesheet" href="/css/styles.css" media="all" />
<link rel="stylesheet" href="/js/fancybox/jquery.fancybox.css" type="text/css" media="screen" />
<link rel="stylesheet" href="/js/fancybox/jquery.fancybox-thumbs.css" type="text/css" media="screen" />


<script type="text/javascript" src="/js/jquery-1.11.2.min.js"></script>
<script type="text/javascript" src="/js/jquery-migrate-1.2.1.min.js"></script>
<script type="text/javascript" src="/js/jquery.easing.1.3.js"></script>
<script type="text/javascript" src="/js/jquery-validate/jquery.validate.js"></script>
<script type="text/javascript" src="/js/jquery-validate/methods.js"></script>
<script type="text/javascript" src="/js/jquery-validate/localization/messages_fr.js"></script>
<script type="text/javascript" src="/js/jquery.placeholder.js"></script>
<script type="text/javascript" src="/js/fancybox/jquery.fancybox.js"></script>
<script type="text/javascript" src="/js/fancybox/jquery.fancybox-thumbs.js"></script>
<script type="text/javascript" src="/js/fancybox/jquery.fancybox-media.js"></script>
<script type="text/javascript" src="/js/jquery.parallax-scroll.js"></script>
<script type="text/javascript" src="/js/communs.js"></script>

<script>
	
	var dateNow = "2024-08-31";
	

	$(document).ready(function() {
		
		function renderCalendar() {
			$('#calendar').fullCalendar({
				header: {
					left: 'prev,next today',
					center: 'title',
					right: 'basicWeek,basicDay'
				},
				defaultDate: dateNow,
				defaultView: 'basicWeek',
				lang: 'fr',
				height: 'auto',
				minTime: '18:00:00',
				maxTime:'21:00:00',
				allDaySlot: false,
				buttonIcons: true, // show the prev/next text
				weekNumbers: false,
				editable: false,
				eventLimit: false, // allow "more" link when too many events
				timeFormat: 'H:mm',
				slotEventOverlap: false,
				weekends: false,
                                viewRender: function(currentView){
                                    var minDate = moment();
                                    if (minDate >= currentView.start && minDate <= currentView.end) {
                                        $(".fc-prev-button").prop('disabled', true); 
                                        $(".fc-prev-button").addClass('fc-state-disabled'); 
                                    }
                                    else {
                                        $(".fc-prev-button").removeClass('fc-state-disabled'); 
                                        $(".fc-prev-button").prop('disabled', false); 
                                    }

                                },
				events: [
																				{
						title: ' ',
						start: '2024-06-03T18:00:00',
						end: '2024-06-03T19:00:00',
												color: '#D3D3D3',
						description: 'Reservé par  '
											},
																				{
						title: 'Legolas L.',
						start: '2024-06-03T18:00:00',
						end: '2024-06-03T19:00:00',
												color: '#D3D3D3',
						description: 'Reservé par Legolas L.'
											},
																				{
						title: 'Gimli G.',
						start: '2024-06-03T18:00:00',
						end: '2024-06-03T19:00:00',
												color: '#D3D3D3',
						description: 'Reservé par Gimli G.'
											},
																				{
						title: 'Aragorn A.',
						start: '2024-06-03T18:00:00',
						end: '2024-06-03T19:00:00',
												color: '#D3D3D3',
						description: 'Reservé par Aragorn A.'
											},
																														{
						title: 'Gandalf G.',
						start: '2024-06-03T19:00:00',
						end: '2024-06-03T20:00:00',
												color: '#D3D3D3',
						description: 'Reservé par Gandalf G.'
											},
																				{
                        title: 'Place disponible',
												color: '#82BCF3',
												url: '?reserver=44036',
						start: '2024-06-03T19:00:00',
						end: '2024-06-03T20:00:00',
						description: 'Cliquez pour réserver entre 19:00 et 20:00'
											},
																				{
						title: 'Sam S.',
						start: '2024-06-03T19:00:00',
						end: '2024-06-03T20:00:00',
												color: '#D3D3D3',
						description: 'Reservé par Sam S.'
											},
																				{
                        title: 'Place disponible',
												color: '#82BCF3',
												url: '?reserver=44036',
						start: '2024-06-03T19:00:00',
						end: '2024-06-03T20:00:00',
						description: 'Cliquez pour réserver entre 19:00 et 20:00'
											},
																														{
						title: 'Frodo F.',
						start: '2024-06-03T20:00:00',
						end: '2024-06-03T21:00:00',
												color: '#D3D3D3',
						description: 'Reservé par Frodo F.'
											},
																				{
						title: 'Place disponible',
												color: '#82BCF3',
												url: '?reserver=44036',
						start: '2024-06-03T20:00:00',
						end: '2024-06-03T21:00:00',
						description: 'Cliquez pour réserver entre 20:00 et 21:00'
											},
																				{
						title: 'Session annulée',
						color: '#BF0000',
						start: '2024-06-03T20:00:00',
						end: '2024-06-03T21:00:00',
						description: 'La session est annulée.'
											},
																				{
						title: 'Session annulée',
						color: '#BF0000',
						start: '2024-06-03T20:00:00',
						end: '2024-06-03T21:00:00',
						description: 'La session est annulée.'
											},
																														]
				eventRender: function(event, element) {
					element.qtip({
						content: event.description,
						 position: {
							 target: 'mouse', // Track the mouse as the positioning target
							 adjust: { x: 5, y: 5 } // Offset it slightly from under the mouse
						 }
					});
				}
			});
		}

		renderCalendar();
	});

</script>
</footer>			
			<a href="#" id="btn_up">
				<img alt="Retour" title="Retour" src="/images/communs/top.png" />
			</a>
			
	</body>
</html>						
"""

calendar_entries_dict = {
    "2024-06-03T18:00:00": "EMPTY,Legolas L.,Gimli G.,Aragorn A.",
    "2024-06-03T19:00:00": "Gandalf G.,FREE,Sam S.,FREE",
    "2024-06-03T20:00:00": "Frodo F.,FREE,CANCELLED,CANCELLED",
}



from datetime import datetime

import requests


def sort_and_format_slots(slots, now=None):
    """
    Sort slots dictionary ({ datetime: count }) by weekday and time.
    Format them as [number_of_weeks_ahead, weekday, time, count]
    """
    sorted_datetimes = sorted(
        [(datetime.fromisoformat(dt), count) for dt, count in slots.items()],
        key=lambda x: x[0].timestamp(),
    )
    current_week = int((now if now is not None else datetime.now()).strftime("%W"))

    return [
        [int(dt.strftime("%W")) - current_week, dt.strftime("%A"), dt.hour, count]
        for dt, count in sorted_datetimes
    ]


def _format_n_slots_text(n):
    if n == 1:
        return "1 slot"
    else:
        return f"{n} slots"


def _format_week_text(n):
    if n == 0:
        return "this week"
    elif n == 1:
        return "next week"
    else:
        return f"in {n} weeks"


def notify_to_telegram(available_slots, token, chat_id):
    text = "*** NEW SLOTS AVAILABLE! ***"

    # Format slot info as [weekday, time, count]. Sort slots by weekday and time

    formatted_slots = sort_and_format_slots(available_slots)

    for [plus_week, day, time, count] in formatted_slots:
        text += f"\n{_format_n_slots_text(count)} available at {time} on {day} {_format_week_text(plus_week)}"

    response = requests.post(
        url=f"https://api.telegram.org/bot{token}/sendMessage",
        data={"chat_id": chat_id, "text": text},
    )

    if response.status_code == 404:
        raise Exception("Wrong bot token or chat ID")



import sqlite3
import re

from ..constants import CANCELLED, FREE, EMPTY, DB_NAME, RUN_WITH_LOGS


def extract_calendar_entries(html):
    # Extract events from the HTML response
    events_text = re.search(r"events\: \[(.*)\]", html, re.DOTALL).group(1)
    trimmed_text = re.sub(r"\n|\t", "", events_text)
    events = re.findall(r"\{.*?\}", trimmed_text)

    """
        We store the calendar entries in a dictionary { [key]: value }, where:
            * key: str
                Date + time of the session in ISO format. E.g.: '2024-08-12T18:00:00'
            * value: str.
                Names of the participants of the session, separated by commas.
                If a session has been cancelled, the name used is 'CANCELLED'
                If a session is free, the name used is 'FREE
                E.g.: 'Bond J.,Norbert E.,FREE,CANCELLED'
    """
    calendar_entries = {}
    for event in events:
        # Events have the formats:
        #   "{title: 'Bond J.',start: '2024-06-03T18:00:00',end: '2024-06-03T19:00:00',color: '#D3D3D3',description: 'Reservé par Bond J.'}"
        #   "{title: ' ',start: '2024-06-03T18:00:00',end: '2024-06-03T19:00:00',color: '#D3D3D3',description: 'Reservé par  '}"
        #   "{title: 'Place disponible',color: '#82BCF3',url: '?reserver=43966',start: '2024-09-18T19:00:00',end: '2024-09-18T20:00:00',description: 'Cliquez pour réserver entre 19:00 et 20:00'}"
        #   "{title: 'Session annulée',color: '#BF0000',start: '2024-09-18T20:00:00',end: '2024-09-18T21:00:00',description: 'La session est annulée.'}"
        name = re.search("title: '(.*?)'", event).group(1)
        if name == "Session annulée":
            name = CANCELLED
        elif name == "Place disponible":
            name = FREE
        elif name == " ":
            name = EMPTY
        datetime = re.search("start: '(.*?)'", event).group(1)
        if datetime in calendar_entries:
            calendar_entries[datetime] += f",{name}"
        else:
            calendar_entries[datetime] = name

    return calendar_entries


def write_calendar_entries_to_db(calendar_entries):
    conn = sqlite3.connect(f"{DB_NAME}")
    cursor = conn.cursor()

    for datetime in calendar_entries:
        cursor.execute(
            """
            SELECT members
            FROM calendar
            WHERE datetime = ?
        """,
            (datetime,),
        )
        old_members_array = cursor.fetchall()

        if len(old_members_array) == 0:
            # No previous entry exists
            cursor.execute(
                """
                INSERT INTO calendar (datetime, members)
                VALUES (?, ?)
            """,
                (datetime, calendar_entries[datetime]),
            )
        else:
            # Overwrite previous entry
            cursor.execute(
                """
                UPDATE calendar
                SET members = ?
                WHERE datetime = ?
            """,
                (calendar_entries[datetime], datetime),
            )

    conn.commit()
    conn.close()


def count_slots_available(new_members: str, old_members: str) -> int:
    new_free_slots = new_members.count(FREE)
    old_free_slots = old_members.count(FREE)

    return max(new_free_slots - old_free_slots, 0)


def find_available_slots(calendar_entries):
    conn = sqlite3.connect(f"{DB_NAME}")
    cursor = conn.cursor()

    result = {}
    has_available_slots = False

    for datetime in calendar_entries:
        cursor.execute(
            """
            SELECT members
            FROM calendar
            WHERE datetime = ?
        """,
            (datetime,),
        )
        old_members_entry = cursor.fetchone()

        if old_members_entry is None:
            # The DB hasn't been populated for this datetime.
            continue
        else:
            old_members = old_members_entry[0]

        if count := count_slots_available(calendar_entries[datetime], old_members):
            has_available_slots = True
            result[datetime] = count

    conn.close()

    if RUN_WITH_LOGS:
        print(
            "* No new slots available :("
            if has_available_slots is False
            else "* Available slots found! :D"
        )

    return result if has_available_slots else None
